/*
(c) Copyright 2008 Pacito.Sys hppacito@gmail.com

┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘

*/

          .file     "vga_refresh.S"

#include "vga_drv_defs.h"

.globl pixel, "o:p"
.type  pixel, @object
          .section .dp.data,       "adw", @progbits
          .align    4
pixel:
          .int      0x00080200
          .align    4
          .section .globcode,      "ax",  @progbits
          .align    2
.LC0:
          ldw       r0, dp[pixel] 
          setc      res[r0], 0x8
          retsp     0x0 
          .section .ctors,         "aw",  @progbits
          .align    4
          .int      .LC0
          .section .globcode,      "ax",  @progbits
.LC1:
          ldw       r0, dp[pixel] 
          setc      res[r0], 0x0
          retsp     0x0 
          .section .dtors,         "aw",  @progbits
          .align    4
          .int      .LC1
.globl vsync, "o:p"
.type  vsync, @object
          .section .dp.data,       "adw", @progbits
          .align    4
vsync:
          .int      0x00010f00
          .align    4
          .section .globcode,      "ax",  @progbits
.LC2:
          ldw       r0, dp[vsync] 
          setc      res[r0], 0x8
          retsp     0x0 
          .section .ctors,         "aw",  @progbits
          .int      .LC2
          .section .globcode,      "ax",  @progbits
.LC3:
          ldw       r0, dp[vsync] 
          setc      res[r0], 0x0
          retsp     0x0 
          .section .dtors,         "aw",  @progbits
          .int      .LC3
.globl hsync, "o:p"
.type  hsync, @object
          .section .dp.data,       "adw", @progbits
          .align    4
hsync:
          .int      0x00010e00
          .align    4
          .section .globcode,      "ax",  @progbits
.LC4:
          ldw       r0, dp[hsync] 
          setc      res[r0], 0x8
          retsp     0x0 
          .section .ctors,         "aw",  @progbits
          .int      .LC4
          .section .globcode,      "ax",  @progbits
.LC5:
          ldw       r0, dp[hsync] 
          setc      res[r0], 0x0
          retsp     0x0 
          .section .dtors,         "aw",  @progbits
          .int      .LC5

          .text
          .align    2
.globl vga_refresh, "f{0}(p(uc))"
.globl vga_refresh.nstackwords
.globl vga_refresh.maxthreads
.globl vga_refresh.maxtimers
.globl vga_refresh.maxchanends
.globl vga_refresh.maxsync
.type  vga_refresh, @function
.linkset vga_refresh.nstackwords, 2
/*
.linkset vga_refresh.locnoside, 0
.linkset vga_refresh.locnochandec, 1
.linkset .LLNK0//, .LLNK1 + 9
.cc_top vga_refresh.function,vga_refresh
*/

          .align    4
vga_refresh:
          entsp     0x2
	  stw	    r0, sp[0x1]  //pointer to buffer
	  ldc	    r4, 0
// outpus default color and sync states
// vsync
          ldc       r5, VSYNCPOL0
          ldw       r3, dp[vsync] 
          out       res[r3], r5 
// hsync
          ldc       r5, VSYNCPOL0
          ldw       r2, dp[hsync] 
          out       res[r2], r5
// pixel data, i.e. color black
	  ldc       r5, 0
          ldw       r11, dp[pixel] 
          out       res[r11], r5 
// we use the 1st timer
          getr      r9, TIMERN 
          ldc       r5, CTRL_COND_NONE // timer will not wait on next IN
          setc      res[r9], r5
          in        r6, res[r9] // gets value of timer (it does not wait)
	  ldc       r0, 1000
	  add       r6, r6, r0 // adds some delay
	  
	  ldc       r5, CTRL_COND_AFTER // timer will wait on next IN
          setc      res[r9], r5
	  
// ethernal loop
vga_ref_ethernal:
          ldc       r0, 0	// Y line counter
	  ldw       r8, sp[0x1] // pixel buffer
	  ldc       r7, 0x0f    // color mask
	  ldaw      r10, dp[palette]
vga_ref_pixellines:
          ldc       r1, 0	// X pixel counter
          ldw       r11, dp[pixel] 
          ldc       r5, XBYTES  // X bytes
	  	  
	  setd      res[r9], r6 // sets new waiting time
          in        r6, res[r9] // waits... gets value of timer (it does not wait)

// r0 : counter of y lines
// r1 : X pixel counter
// r2 : <scratch>
// r3 : <scratch>
// r4 : 0
// r5 : XCOLS
// r6 : hsync_end
// r7 : 0x0f
// r8 : pointer to bitmap buffer
// r9 : timer
//r10 : palette pointer
//r11 : pixel (port)

// each instruction needs 4 ticks, so we have 100 MIPS 
vga_ref_xpixels:
          ld8u      r3, r8[r1]	// reads pixel data
	  shr       r2, r3, 4
	  ld8u	    r2, r10[r2] // gets color data from palette
          out       res[r11], r2
	  nop
	  nop
	  nop
	  nop
	  nop
	  and       r3, r3, r7
	  ld8u	    r2, r10[r3] // gets color data from palette
          out       res[r11], r2
// 1 nop to compensate
          nop
          add       r1, r1, 0x1
          lss       r3, r1, r5
          bt        r3, vga_ref_xpixels

// sets color output to 0 
	  nop
	  nop
	  nop
          out       res[r11], r4 

          ldc       r2, 0x01
	  and       r2, r2, r0
          bf        r2, vga_ref_noadd // repeats line
	  
	  add       r8, r8, r1	// adds 160 to pointer, so next line 
vga_ref_noadd:

// Starting here
// r0 : counter of y lines
// r1 : <scratch>
// r2 : <scratch>, later hsync (port)
// r3 : <scratch>, later vsync (port)
// r4 : 0
// r5 : <scratch>
// r6 : hsync_end
// r7 : bitmap_buffer
// r8 : pointer to bitmap buffer
// r9 : timer
//r10 : palette pointer
//r11 : pixel (port)
	  
          ldw       r2, dp[hsync] 


// waits for timer to let elapse front porch
          ldc       r1, HSYNCSTART
	  add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9]
// starts hsync pulse 
          ldc       r1,HSYNCPOL1
          out       res[r2], r1
// waits for timer
          ldc       r1, HSYNCEND
          add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9] 
// ends sync pulse
          ldc       r1,HSYNCPOL0
          out       res[r2], r1
// starts back porch, waits for timer (at the beggining of the loop)
          ldc       r1, HSYNCBP
          add       r6, r6, r1

// increments line counter and compares
          add       r0, r0, 0x1
	  ldc       r1, RYLINES // numer of real Y lines
          lss       r1, r0, r1
          bt        r1, vga_ref_pixellines
	  
// clears front porch counter 
          mov       r0, r4
vga_ref_front_porch:

          ldc       r1, HSYNCSTART
	  add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9]
// starts hsync pulse 
          ldc       r1, HSYNCPOL1
          out       res[r2], r1 
// waits for timer
          ldc       r1, HSYNCEND
          add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9] 
// ends sync pulse
          ldc       r1, HSYNCPOL0
          out       res[r2], r1
// starts back porch, waits for timer
          ldc       r1, HSYNCBP
          add       r6, r6, r1
          setd      res[r9], r6
          in        r6, res[r9]

          add       r0, r0, 0x1
          ldc       r1, 10 
          lss       r1, r0, r1
          bt        r1, vga_ref_front_porch

// starts vsync 
          ldc       r1, VSYNCPOL1
          ldw       r3, dp[vsync] 
          out       res[r3], r1 
// clears front porch line counter
          mov       r0, r4
vga_ref_vsync:

          ldc       r1, HSYNCSTART
	  add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9]
// starts hsync pulse 
          ldc       r1, HSYNCPOL1
          out       res[r2], r1 
// waits for timer
          ldc       r1, HSYNCEND
          add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9] 
// ends sync pulse
          ldc       r1, HSYNCPOL0
          out       res[r2], r1
// starts back porch, waits for timer
          ldc       r1, HSYNCBP
          add       r6, r6, r1
          setd      res[r9], r6
          in        r6, res[r9]

          add       r0, r0, 0x1
          ldc       r1, 2
          lss       r1, r0, r1
          bt        r1, vga_ref_vsync 
	  
	  
// ends vsync pulse	  
          ldc       r1, VSYNCPOL0
          out       res[r3], r1
	  
// sets to 0 counter of back porch lines
          mov       r0, r4

vga_ref_back_porch:

          ldc       r1, HSYNCSTART
	  add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9]
// starts hsync pulse 
          ldc       r1, HSYNCPOL1
          out       res[r2], r1
// waits for timer
          ldc       r1, HSYNCEND
          add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9] 
// ends sync pulse
          ldc       r1, HSYNCPOL0
          out       res[r2], r1
// starts back porch, waits for timer
          ldc       r1, HSYNCBP
          add       r6, r6, r1
          setd      res[r9], r6
          in        r6, res[r9]

          add       r0, r0, 0x1
          ldc       r1, 32		// back porch lines
          lss       r1, r0, r1
          bt        r1, vga_ref_back_porch 
	  
// one more line
          ldc       r1, HSYNCSTART
	  add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9]
// starts hsync pulse 
          ldc       r1, HSYNCPOL1
          out       res[r2], r1
// waits for timer
          ldc       r1, HSYNCEND
          add       r6, r6, r1
          setd      res[r9], r6
          in        r1, res[r9] 
// ends sync pulse
          ldc       r1, HSYNCPOL0
          out       res[r2], r1
// starts back porch, waits for timer
          ldc       r1, HSYNCBP
          add       r6, r6, r1
          
	  bu        vga_ref_ethernal

.globl palette, "o:p"
.type  palette, @object
          .section .dp.data,       "adw", @progbits
          .align    4
palette:
          .byte      0
	  .byte	     0x20	// blue
	  .byte	     0x28	// cyan
	  .byte	     0x08	// green
	  .byte      0x0a	// yellow
	  .byte      0x02	// red
          .byte      0x15	// dark grey
	  .byte      0x2c	// medium grey
	  .byte      0x3f	// light gray
	  .byte      0x30	// light blue
	  .byte	     0x3c	// light cyan
	  .byte      0x0c	// light green
	  .byte      0x0f	// light yellow
	  .byte	     0x03	// light red
	  .byte      0x01       // dark red
	  .byte      0x04       // dark green
